'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

exports.mergerware = mergerware;

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

function flatten(arg) {
  return arg.reduce(function (memo, item) {
    return Array.isArray(item) ? [].concat(_toConsumableArray(memo), _toConsumableArray(flatten(item))) : [].concat(_toConsumableArray(memo), [item]);
  }, []);
}

function normalizeArgumentArray(args) {
  return flatten(args).filter(function (item) {
    return !!item;
  });
}

function isFunction(fn) {
  return typeof fn === 'function';
}

function noop() {}

function isPromise(obj) {
  return !!obj && ((typeof obj === 'undefined' ? 'undefined' : _typeof(obj)) === 'object' || typeof obj === 'function') && typeof obj.then === 'function';
}

function mergerware(merger) {
  var initial = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

  // high level - returns a function that can merge results
  return function middle() {
    for (var _len = arguments.length, middleware = Array(_len), _key = 0; _key < _len; _key++) {
      middleware[_key] = arguments[_key];
    }

    // this is the actual warewolf() call
    var argStack = normalizeArgumentArray(middleware);
    var hasCallback = true;
    var isDone = false;
    return function () {
      for (var _len2 = arguments.length, invocationArguments = Array(_len2), _key2 = 0; _key2 < _len2; _key2++) {
        invocationArguments[_key2] = arguments[_key2];
      }

      // some assumptions here about how warewolf is called
      // if callback is provided, that's subbed with next() in middleware
      // if no callback, we treat it like a promise
      var done = invocationArguments.pop();
      if (!isFunction(done)) {
        if (done) {
          invocationArguments.push(done);
        }
        done = noop;
        hasCallback = false;
      }
      var stack = [].concat(_toConsumableArray(argStack));
      // used with merger
      var model = initial;
      return new Promise(function (resolve, reject) {
        try {
          // wrapping execution with try catch so we can properly reject
          var wrappedMiddlewareStep = function wrappedMiddlewareStep(middlewareStep) {
            // wrapping middlewareStep with some functionality
            var stepArguments = merger !== undefined && stack.length === 0 ? [].concat(invocationArguments, [model]) : invocationArguments;

            var middlewareStepCallback = function middlewareStepCallback(err) {
              for (var _len3 = arguments.length, result = Array(_len3 > 1 ? _len3 - 1 : 0), _key3 = 1; _key3 < _len3; _key3++) {
                result[_key3 - 1] = arguments[_key3];
              }

              if (err || stack.length === 0) {
                if (err) {
                  if (!hasCallback) {
                    // avoiding unhandled rejection error
                    reject(err);
                  }
                } else {
                  resolve(result[0]);
                }
                isDone = true;
                return done.apply(undefined, [err].concat(result));
              }
              if (merger !== undefined) {
                model = merger.apply(undefined, [model].concat(result, invocationArguments));
              }
              var nextStep = stack.shift();
              return wrappedMiddlewareStep(nextStep);
            };

            var result = middlewareStep.apply(undefined, _toConsumableArray(stepArguments).concat([middlewareStepCallback]));
            if (result !== undefined && isPromise(result)) {
              return result.then(middlewareStepCallback);
            }
            return result;
          };
          wrappedMiddlewareStep(stack.shift());
        } catch (err) {
          if (hasCallback) {
            if (!isDone) {
              // want to avoid stack overflow
              done(err);
            } else {
              throw err;
            }
          } else {
            reject(err);
          }
        }
      });
    };
  };
}

exports.default = mergerware();